open Grid
open Commands
open Util
open Parser
open Helpers
open Unix


let myId = ref 0 (* we get this from the server; it's constant *)
let maxCap = ref 0 (* ditto *)

let infiniteWeight = 10000005

let clearString = "\027\091\072\027\091\074\n";;

(***************************************************************)
pa_ocamllex let ws = [' ' '\013' '\009' '\012' ]
pa_ocamllex let digit = ['0'-'9']

pa_ocamllex rule lexServerResp = parse
  ws+ { lexServerResp lexbuf }
| '#'digit+ { let id = Lexing.lexeme lexbuf in (RobotId (int_of_string (String.sub id 1 ((String.length id) - 1)))) }
| ['N''S''E''W' ] { Direction (string_to_direction(Lexing.lexeme lexbuf)) }
| ['X'] { X }
| ['Y'] { Y }
| ['P'] { PickT }
| ['D'] { DropT }
| digit+      { Coord (int_of_string (Lexing.lexeme lexbuf)) }
| eof { EOF }
(***************************************************************)
let char_to_space = function
     '.' -> Plain
  |  '~' -> Water
  |  '#' -> Wall
  |  '@' -> HomeBase
  |  _ -> raise Util.Fail("char_to_space")

let parseBoard (board: string):grid = let rows = (List.map Util.explode (lines board)) in
    Array.init (List.length rows) (fun rowNum -> Array.init (List.length (List.hd rows)) (fun colNum -> ([], [], (char_to_space (List.nth (List.nth rows rowNum) colNum)))))


let rec parsePackages (packageString: string): package list = (List.map stringsToPackage (group 4 (words packageString)))

and stringsToPackage (package: string list):package =
      match package with
         (pkgId::(destX::(destY::(weight::[])))) -> (int_of_string(pkgId), (None, (int_of_string(destY) - 1,int_of_string(destX) - 1), int_of_string(weight)))
       | _ -> raise Util.Fail "stringsToPackage: bad package description"

and getPackages inSock = input_line inSock

and determineCommand (money: int) (pkgs: package list) (gameState: gameState):command =
  Strategy.runStep money pkgs gameState !myId !maxCap

and getResponse (inSock: in_channel) (outSock: out_channel) (command:string):string = 
 (* send command to server and return the response *)
 print_endline ("sending " ^ command ^ " to the server ");
 flush_all();
 output_string outSock (command ^ "\n");
 flush outSock;
 print_endline ("reading response from server");
 flush_all();
 let res = input_line inSock in
 print_endline ("done! response = " ^ res);
 res
       
and updateGameState (gameState:gameState) (serverResponse:string):gameState = 
   let events = parseServerResponse serverResponse in
     print_endline ("parsed server response: " ^ (eventList_to_string events) ^ " end server response"); 
     let res = List.fold_left (fun intermedState event -> applyEventToGameState event intermedState) gameState events in
       res

and parseServerResponse(serverResponse:string):event list =
(* important: need to subtract 1 from all coordinates in moves, b/c the server starts at (1, 1) and we start at (0, 0) *)
(* also exchange x with y, b/c they use row-major and we use column-major or something *)
     print_endline( "parseServerResponse: uninterpreted response = " ^ serverResponse); (List.map subOne (Parser.main lexServerResp (Lexing.from_string serverResponse)))

and subOne ((r, commandType):event):event =
  (r, (match commandType with
        Appear ((x, y), dirs) -> Appear ((y-1, x-1),dirs)
      | foo -> foo))

and applyEventToGameState ((robotId, command): event) ((grid, robots): gameState):gameState = 
 ((* print_endline *) (fun s -> ()) ("applyEventToGameState: command = " ^ (command_to_string (robotId, command))); 
  match command with
     Appear ((x, y), directions) -> (* need to find existing robot in the list (if there is one), to preserve packages *)    
        
        (try (let (((oldx, oldy) as oldPos), pkgs) = (List.assoc robotId robots) in
               let (newx, newy) as newPos = goDirections (x, y) directions in
                (* in robot list: update the position in the robot *)
                let newRobots = ((robotId, (newPos, pkgs))::(List.remove_assoc robotId robots)) in
                  (* in grid: remove the robot from old pos, along with any of its packages *)
                  let newGrid = deepCopy grid in
                    let (oldRobots, oldPkgs, spaceTy) = newGrid.(oldx).(oldy) in
                     let (oldRobotsInNewSpace, oldPkgsInNewSpace, newSpaceTy) = newGrid.(newx).(newy) in
                      let robotsMinusOne = remove robotId oldRobots in
                       let newPkgs = (List.fold_right (fun pkgId spacePkgs -> List.remove_assoc pkgId spacePkgs) pkgs oldPkgs) in
(* Note: we know that we're assigning to the array within its bounds, b/c we're dealing with data from the server. *)
                          newGrid.(oldx).(oldy) <- (robotsMinusOne, newPkgs, spaceTy);
(* add the robot and its packages to the new space *)
                          let myPkgs = (List.map (fun pkgId -> (pkgId, (try List.assoc pkgId oldPkgs with Not_found -> (None, (-1, -1), infiniteWeight)))) pkgs) in
                          newGrid.(newx).(newy) <- ((addNoDups robotId oldRobotsInNewSpace), myPkgs@oldPkgsInNewSpace, newSpaceTy); 
                          (* print_endline ("applyEventToGameState: added robot at " ^ (string_of_int newx) ^ " " ^ (string_of_int newy));
                          print_endline ("applyEventToGameState: oldRobots = " ^ (robotList_to_string !myId robots));
                          print_endline ("applyEventToGameState: newRobots = " ^ (robotList_to_string !myId newRobots) ^ "*****"); *)
                          (newGrid, newRobots)) 
         with Not_found -> (* it's a new robot *)
                             print_endline ((string_of_int robotId) ^ " is a new robot -- not in  " ^ (robotList_to_string !myId robots));
                             (let (newx, newy) as newPos = goDirections (x, y) directions in
                               let newRobots = (robotId, (newPos, []))::robots in
                                let newGrid = deepCopy grid in
                                 let (oldRobots, oldPkgs, spaceTy) = newGrid.(newx).(newy) in
                                   newGrid.(newx).(newy) <- (robotId::oldRobots, oldPkgs, spaceTy);
                                   (newGrid, newRobots)))
(* a move just desugars to an Appear at the current space *)
|    Move directions -> (applyEventToGameState (robotId, (Appear ((getRobotPos robotId robots), directions))) (grid, robots))
|    Pick pkgs -> (let ((x, y) as pos, oldPkgs) = (List.assoc robotId robots) in
                    (* first, update robot list *)
(* Q: what happens if you pick up a package you already have? Assuming this won't happen. *)
                     let newRobots = (robotId, (pos, pkgs@oldPkgs))::(List.remove_assoc robotId robots) in
                    (* then, make packages "owned" in the space *)
                       let newGrid = deepCopy grid in
                         let (spaceRobots, oldPkgs, spaceTy) = newGrid.(x).(y) in
                           let (pickedPkgs, unpickedPkgs) = List.partition (fun (p,_) -> (List.mem p pkgs)) oldPkgs in
                             let ownedPkgs = (List.map (makeOwnedBy robotId) pickedPkgs) in
                               print_endline ("applyEventToGrid: oldPkgs = " ^ (packageList_to_string oldPkgs));
                               print_endline ("applyEventToGrid: picked pkgs = " ^ (packageList_to_string pickedPkgs) ^ " unpicked pkgs = " ^ (packageList_to_string unpickedPkgs));
                               print_endline ("applyEventToGrid: adding " ^ (packageList_to_string ownedPkgs) ^ " to " ^ (string_of_int x) ^ " " ^ (string_of_int y));
                               print_endline ("applyEventToGrid: setting packages to " ^ (packageList_to_string (ownedPkgs@unpickedPkgs)));
                               newGrid.(x).(y) <- (spaceRobots, ownedPkgs@unpickedPkgs, spaceTy);
                               (newGrid, newRobots))  
|    Drop pkgs -> let ((x, y) as pos, myoldPkgs) = (List.assoc robotId robots) in
                    (* first, update robot list *)
(* Q: what happens if you pick up a package you already have? Assuming this won't happen. *)
                     let newRobots = ((robotId, (pos, ((List.fold_right (fun pkg rest -> remove pkg rest) pkgs myoldPkgs)))))::(List.remove_assoc robotId robots) in
                    (* then, make packages un"owned" in the space *)
                       let newGrid = deepCopy grid in
                         let (spaceRobots, oldPkgs, spaceTy) = newGrid.(x).(y) in
                           let (droppedPkgs, undroppedPkgs) = List.partition (fun (p,_) -> (List.mem p pkgs)) oldPkgs in
(* Need to figure out whether packages are being delivered! If so, they disappear. *)
                             let unownedPkgs = (List.map makeUnowned droppedPkgs) in
                               let notDelivered = (List.filter (fun (id, (_, (destX, destY), _)) -> (destX != x) || (destY != y)) unownedPkgs) in
                               print_endline ("applyEventToGrid: removed " ^ (packageIdList_to_string pkgs) ^ " from " ^ (packageIdList_to_string myoldPkgs) ^ " result robotlist = " ^ (robotList_to_string !myId newRobots));
                               print_endline ("applyEventToGrid: notDelivered = " ^ (packageList_to_string notDelivered));
                               print_endline ("applyEventToGrid: current pos = " ^ (string_of_int x) ^ " " ^ (string_of_int y));
                               newGrid.(x).(y) <- (spaceRobots, notDelivered@undroppedPkgs, spaceTy);
                               (newGrid, newRobots)) 


(* note: when we print out coordinates, it's with respect to (0, 0) as the origin *)
and gameState_to_string ((grid, robots):gameState):string = 
    (String.concat "\n" (Array.to_list (Array.map row_to_string grid))) 
  ^ "\n" 
  ^ (String.concat "\n" (List.map (robot_to_string !myId) robots))
  ^ "\npackages at current position: " ^ (packageList_to_string (Helpers.getPackages (robotLocation !myId (grid, robots)) (grid, robots)))

and row_to_string (row: space array):string = String.concat "" (Array.to_list (Array.map space_to_string row))

and space_to_string ((robots, packages, spaceType):space):string = 
   if (List.mem !myId robots) then
      "*"
   else (match robots with
        | [] -> (match spaceType with
                      | Plain -> "."
                      | Water -> "~"
                      | Wall -> "#"
                      | HomeBase -> "@")
        | (robId::_) -> (string_of_int (robId mod 10)))  

and addPackages (pkgs: package list) (((grid, robots) as g):gameState): gameState =
(* given a list of packages at my current position, add those to the space *)

(* Suppose there is a 10000005-weight package at the current location,
   and its id is not in the list from the server.
   Then we know that this package must have been delivered, so we erase it. *)
  let newGrid = deepCopy grid in
    let (myX, myY) = robotLocation !myId g in
      let (_, newp, _) as newSpace = (match newGrid.(myX).(myY) with
                        (robots, oldPkgs, ty) -> 
                          let pkgsToBeAddedTo = (List.filter (fun (id, (_, _, weight)) -> (weight != infiniteWeight) || List.mem_assoc id pkgs) oldPkgs) in
                             print_endline ("pkgsToBeAddedTo = " ^ (packageList_to_string pkgsToBeAddedTo));
                             (robots, (addListNoDups_assoc pkgs pkgsToBeAddedTo), ty)) in
        print_endline ("addPackages: setting packages to " ^ (packageList_to_string newp));
        newGrid.(myX).(myY) <- newSpace;
        (newGrid, robots)
                      
and loop (inSock: in_channel) (outSock: out_channel) (money: int) (gameState: gameState): unit = 
  (* get packages from server *)
   try
  let str = getPackages inSock in
   print_endline ("pkg string = " ^ str);
   flush_all();
   let pkgs = parsePackages str in
  (* determine the next command to execute *)
    print_endline ("pkgs = " ^ (packageList_to_string pkgs));
    flush_all();
    let newGameState = addPackages pkgs gameState in
    let command = determineCommand money pkgs newGameState in
  (* how much money do we have after executing this command? *)
    let newMoney = money - (bid command) in
  (* submit command to server and get a response *)
      print_endline ("calling getResponse on " ^ (command_to_string command));
      flush_all();
      let update = getResponse inSock outSock (command_to_string command) in
       flush_all();
       print_endline ("server response = " ^ update);
       flush_all();
  (* update the gameState with the new info *)
       let nextGameState = updateGameState newGameState update 
    in
      print_endline(clearString ^(gameState_to_string(nextGameState)));
      loop inSock outSock newMoney nextGameState 
   with Failure("lexing: empty token") -> die inSock money
      | Util.Fail -> die inSock money

and die (inSock: in_channel) (money:int)  :unit = 
(* try to figure out why *)
   let serverResp = ref "" in
     (print_endline("Money = " ^ (string_of_int money) ^ "\nTe morituro saluto."); 
      try
        while(true) do
          serverResp := !serverResp ^ "\n" ^ (input_line inSock)
        done
      with End_of_file -> print_endline(!serverResp); flush_all(); ())


let readDimensionsFromServer (inSock: in_channel):point =
    let nums = (words (input_line inSock)) in
     (int_of_string(List.hd nums), int_of_string(List.hd (List.tl nums))) 

let readBoardFromServer ((numCols, numRows):point) (inSock: in_channel): string =
  let result = ref "" in
    for i=1 to numRows do
      result := ((!result)^"\n"^(input_line inSock))
    done;
    !result

let readInfoFromServer (inSock: in_channel):(int * int * int) =
  let nums = words (input_line inSock) in
    (int_of_string (List.nth nums 0), int_of_string (List.nth nums 1), int_of_string (List.nth nums 2))

let () = begin
           if Array.length Sys.argv < 3 
             then print_endline "usage :  client server port\n"
            else
(* open connection to server *)
              let server = Sys.argv.(1) in
                 let server_addr =
                   try  Unix.inet_addr_of_string server 
                   with Failure("inet_addr_of_string") -> 
                          try  (Unix.gethostbyname server).Unix.h_addr_list.(0) 
                          with Not_found ->
                                 Printf.eprintf "%s : Unknown server\n" server ;
                                 exit 2
                 in try 
                      let port = int_of_string (Sys.argv.(2)) in
                      let sockaddr = Unix.ADDR_INET(server_addr,port) in 
                        print_endline ("about to open connection");
                        flush_all();
                      let (inSock, outSock) = open_connection sockaddr in
     (* receive board and initial config from server, eventually *)
                        print_endline ("opened connection; sending player");
                        flush_all();
                        output_string outSock "Player\n";
                        flush outSock;
                        print_endline ("reading dimensions from server");
                        flush_all();
                        let (boardWidth, boardHeight) = readDimensionsFromServer inSock in
                          print_endline ("width = " ^ (string_of_int boardWidth) ^ " height = " ^ (string_of_int boardHeight));
                          let board = readBoardFromServer (boardWidth, boardHeight) inSock in
                         (*   print_endline ("board = " ^ board); *)
                            let (id, cap, initMoney) = readInfoFromServer inSock in
                            myId := id;
                            maxCap := cap;
                            print_endline ("myId = " ^ (string_of_int !myId) ^ " maxCap = " ^ (string_of_int !maxCap) ^ " initMoney = " ^ (string_of_int initMoney));
                            let initialEvents = input_line inSock in
                              let gameState = updateGameState ((parseBoard board), []) initialEvents in                                              print_endline ((gameState_to_string gameState));
                            flush_all();
                          loop inSock outSock initMoney gameState;
                          shutdown_connection inSock;
                          exit 0
                    with Failure("int_of_string") -> Printf.eprintf "bad port number";
                                                     exit 2
         end
