open Commands
open Helpers
open Grid

let myId = ref 0;;
let maxCap = ref 0;;
let exhaustedBases = ref [];;

let turnsFighting = ref 0;;
let currentBid = ref 1;;

type mode = Offensive | Defensive;;

type moveMode = Pathfinding | Fighting of point;;

let dist (x1, y1) (x2, y2) =
  (abs (x1-x2)) + (abs (y1-y2))
;;

(* Pathfinding *)

type node = int * int;;

type nodeInfo = int * int * int * int;; (* N,S,E,W *)

type graph = (nodeInfo array) array;;

type spNode = { parent: node; dist: int };;

let weightsToGraph weights excluded =
  let width = Array.length weights in
  let height = Array.length weights.(0) in
  let access x y =
    if x >= 0 && y >= 0 && x < width && y < height then weights.(x).(y)
    else 0
  in
  let is_excluded x y =
    if x >= 0 && y >= 0 && x < width && y < height then excluded.(x).(y)
    else true
  in
  let map val1 val2 =
    let diff = val1 - val2 in
    if diff = 0 then 3
    else if diff > 5 then 1
    else if diff > 0 then 2
    else if diff < -5 then 15
    else 5
  in
  let graph = Array.make_matrix width height (0, 0, 0, 0) in
  for x = 0 to width-1 do
    for y = 0 to height-1 do
      if is_excluded x y then
	graph.(x).(y) <- (-1, -1, -1, -1)
      else
	let self = access x y in
	let north = if is_excluded x (y+1) then -1 else map (access x (y+1)) self in
	let south = if is_excluded x (y-1) then -1 else map (access x (y-1)) self in
	let east = if is_excluded (x+1) y then -1 else map (access (x+1) y) self in
	let west = if is_excluded (x-1) y then -1 else map (access (x-1) y) self in
	let edges = (north, south, east, west) in
	graph.(x).(y) <- edges
    done
  done;
  graph
;;

let debug x y =
  () (*
  print_string "debug "; print_int x; print_string " "; print_int y; print_string "\n" *)
;;

let printPathInfo info =
  let width = Array.length info in
  let height = Array.length info.(0) in
  for y = 0 to height-1 do
    for x = 0 to width-1 do
      let n = info.(x).(y) in
      let (px, py) = n.parent in
      Printf.printf "[%d (%d,%d)] " n.dist px py
    done;
    print_string "\n"
  done
;;

let findPath graph (sx, sy) =
  let width = Array.length graph in
  let height = Array.length graph.(0) in
  let infinity = 100000 in
  let info = Array.make_matrix width height {parent = (0, 0); dist = infinity} in
  info.(sx).(sy) <- {parent = (0,0); dist = 0};
  let find_min (mdist, mnode, mparent) (x, y) =
    let node = info.(x).(y) in
    if node.dist < mdist then (node.dist, (x, y), node.parent)
    else (mdist, mnode, mparent)
  in
  let rec run_algo scomp =
    let (mdist, mnode, mparent) = List.fold_left find_min (infinity, (0, 0), (0, 0)) scomp in
    if mdist = infinity then () else
    let scomp' = ref (List.filter (fun n -> n <> mnode) scomp) in
    let (x,y) = mnode in
    let sdist = info.(x).(y).dist in
    let (north, south, east, west) = graph.(x).(y) in 
    let relax (rx, ry) edgeinfo =
      if info.(rx).(ry).dist > sdist + edgeinfo then
	begin
	  if info.(rx).(ry).dist = infinity then scomp' := (rx, ry) :: (!scomp');
	  info.(rx).(ry) <- { dist = sdist + edgeinfo; parent = mnode }
	end
    in
    if (north >= 0) then relax (x, y+1) north;
    if (south >= 0) then relax (x, y-1) south;
    if (east >= 0) then relax (x+1, y) east;
    if (west >= 0) then relax (x-1, y) west;
    run_algo (!scomp')
  in
  let rec genNodes x y rest =
    let newRest = (x, y) :: rest in
    if x = width-1 && y = height-1 then newRest
    else if x = width-1 then genNodes 0 (y+1) newRest
    else genNodes (x+1) y newRest
  in
  run_algo [(sx, sy)];
  info
;;

let find_path_distance info (x, y) =
  info.(x).(y).dist
;;

let rec find_path_path info (x, y) =
  if info.(x).(y).dist = 0 then [(x, y)]
  else (x, y) :: find_path_path info (info.(x).(y).parent)
;;

(* Pickup code *)

type merge_type =
    Carried
  | Available
;;

let pickupStrategy othersAround position carried available capacity =
  let value (id, (robot, ppos, weight)) =
    let lambda = 1.0 in
    (float (weight)) -. lambda *. (float (dist ppos position))
  in
  let greedyPickup () =
    let rec doGreedy space toPickup = function
	[] -> toPickup
      |	(id, (robot, ppos, weight)) :: tail ->
	  if weight <= space then
	    doGreedy (space - weight) (id :: toPickup) tail
	  else
	    doGreedy space toPickup tail
    in
    let cmp x y =
      if (value x) > (value y) then -1
      else if (value y) = (value x) then 0
      else 1
    in
    let savail = List.sort cmp available in
    let carrying = List.fold_left (fun cur (id, (r, p, w)) -> cur + w) 0 carried in
    let space = capacity - carrying in
    doGreedy space [] savail
  in
  let knapsackPickup () =
    let rec doIteration space toPickup toDrop = function
	(typ, (id, (robot, ppos, weight))) :: tail ->
	  if weight <= space then
	    if typ = Available then doIteration (space - weight) (id :: toPickup) toDrop tail
	    else doIteration (space - weight) toPickup toDrop tail
	  else
	    if typ = Carried then doIteration space toPickup (id :: toDrop) tail
	    else doIteration space toPickup toDrop tail
      |	[] ->
	  (toDrop, toPickup)
    in
    let carriedMerged = List.map (fun x -> (Carried, x)) carried in
    let availMerged = List.map (fun x -> (Available, x)) available in
    let merged = carriedMerged @ availMerged in
    let cmp (a, x) (b, y) =
      if (value y) > (value x) then 1
      else if (value y) = (value x) then 0
      else -1
    in
    let smerged = List.sort cmp merged in
    doIteration capacity [] [] smerged
  in
  if othersAround then
    let result = greedyPickup () in
    if result = [] then None else Some(!currentBid, Pick(result))
  else
    match knapsackPickup () with
      ([], []) -> None
    | (toDrop, []) -> Some(1, Drop(toDrop))
    | ([], toPickup) -> Some(1, Pick(toPickup))
    | (toDrop, toPickup) -> Some(1, Drop(toDrop))
;;

(** Testing **)

let rec join = function
    e1 :: e2 :: tail ->
      e1 ^ ", " ^ (join (e2 :: tail))
  | e1 :: [] ->
      e1
  | [] ->
      ""
;;

let printGraph graph =
  let width = Array.length graph in
  let height = Array.length graph.(0) in
  for x = 0 to width-1 do
    for y = 0 to height-1 do
      let (n, s, e, w) = graph.(x).(y) in
      Printf.printf "[%d:%d:%d:%d]" n s e w
    done;
    print_string "\n"
  done
;;

(* Main code *)

type weightAssignment =
    { nearbyWater : int;
      package : int;
      robotPackage : int;
      robot : int;
      delivery : int
    };;

let defensiveWeights =
  { nearbyWater = -5;
    package = 2;
    robotPackage = -2;
    robot = -8;
    delivery = 10
  };;

let offensiveWeights =
  { nearbyWater = -1;
    package = 10;
    robotPackage = 5;
    robot = -2;
    delivery = 4
  };;

let debugBoard gameState =
  let (width, height) = bounds gameState in
  for x = 0 to width-1 do
    for y = 0 to height-1 do
      match getType (x,y) gameState with
	Water -> print_char '~'
      |	Wall -> print_char '#'
      |	Plain -> print_char '.'
      |	HomeBase -> print_char '@'
    done;
    print_char '\n'
  done
;;

let computeGraph pkgs gameState mode =
  let isExcluded (x, y) = 
    match getType (x,y) gameState with 
      Water -> true
    | Wall -> true
    | Plain -> false
    | HomeBase -> false
  in let checkSurroundings (x, y) typ =
    if getType (x, y+1) gameState = typ then true
    else if getType (x, y-1) gameState = typ then true
    else if getType (x-1, y) gameState = typ then true
    else if getType (x+1, y) gameState = typ then true
    else false
  in let getWeight (x, y) =
    let myPkgs = robotPackages !myId gameState in
    let allPkgs = getPackages (x, y) gameState in
    let pkgs = List.filter (fun p -> (packageOwner p) <> Some(!myId)) allPkgs in
    let allRobots = getRobots (x, y) gameState in
    let robots = List.filter (fun r -> r <> !myId) allRobots in
    let position = robotLocation !myId gameState in
    let assignWeight values =
      if checkSurroundings (x, y) Water then values.nearbyWater
      else if List.exists (fun p -> isOwned p) pkgs then
	values.robotPackage
      else if pkgs <> [] then values.package
      else if List.exists (fun p -> packageDest p position gameState = (x, y)) myPkgs then
	values.delivery
      else if robots <> [] then values.robot
      else 0
    in
    match mode with
      Defensive -> assignWeight defensiveWeights
    | Offensive -> assignWeight offensiveWeights
  in
  let (width, height) = bounds gameState in
  let weights = Array.make_matrix width height 0 in
  let excluded = Array.make_matrix width height false in
  for x = 0 to width-1 do
    for y = 0 to height-1 do
      excluded.(x).(y) <- isExcluded (x, y);
      weights.(x).(y) <- getWeight (x, y);
      (* Printf.printf "[%d]" weights.(x).(y) *)
    done
    (* print_char '\n' *)
  done;

  let result = weightsToGraph weights excluded in
  result
;;

let findGoal (money: int) (pkgs: package list) (gameState: gameState) (mode: mode) info:point =
  let infinity = 1000000 in
  let position = robotLocation !myId gameState in
  let (width, height) = bounds gameState in
  let findDeliver () =
    let myPkgs = robotPackages !myId gameState in
    let minimizer (min, minp) p =
      let dest = packageDest p position gameState in
      let dist = find_path_distance info dest in
      if dist < min then (dist, dest) else (min, minp)
    in
    let (min, minp) = List.fold_left minimizer (infinity, (0, 0)) myPkgs in
    if min = infinity then None else Some(minp)
  in let rec findPickup (x, y) (min, minp) =
    if y >= height then minp else
    let nextp = if x = width-1 then (0, y+1) else (x+1, y) in
    let typ = getType (x, y) gameState in
    let pkgs = getPackages (x, y) gameState in
    let dist = find_path_distance info (x, y) in
    let base = typ = HomeBase && not (List.mem (x, y) !exhaustedBases)in
    if base && position = (x, y) then
      exhaustedBases := (x, y) :: !exhaustedBases;
    if (pkgs <> [] || base) && dist < min then findPickup nextp (dist, (x, y))
    else findPickup nextp (min, minp)
  in
  match findDeliver () with
    None -> findPickup (0, 0) (infinity, (0, 0))
  | Some(p) -> p
;;

let findNextTile info dest =
  let rec findPenultimate = function
      [] -> raise Util.Fail
    | one :: [] -> one
    | two :: one :: [] -> two
    | head :: tail -> findPenultimate tail
  in
  let path = find_path_path info dest in
  findPenultimate path
;;

let checkOthersAround (x, y) gameState =
  if getRobots (x, y+1) gameState <> [] then true
  else if getRobots (x, y-1) gameState <> [] then true
  else if getRobots (x-1, y) gameState <> [] then true
  else if getRobots (x+1, y) gameState <> [] then true
  else false
;;

let findOthersAround gameState info thresh =
  let (width, height) = bounds gameState in
  let rec iterate (x, y) =
    if y = height then None
    else if x = width then iterate (0, (y+1))
    else
      let allRobots = getRobots (x, y) gameState in
      let robots = List.filter (fun r -> r <> !myId) allRobots in
      if robots <> [] && ((find_path_distance info (x, y)) <= thresh) then Some((x, y))
      else iterate (x+1, y)
  in
  iterate (0, 0)
;;

exception Overflow;;

let baseConfig = [((N, N), 1); ((N, N), 1); ((N, N), 1)];;
let baseTurn = ((N, N), 1);;
let baseMoves = (N, N);;
let baseMove = N;;

let maxOrder = 2;;

let incrementMove = function
    N -> S
  | S -> E
  | E -> W
  | W -> raise Overflow
;;

let rec incrementMoves (m1, m2) =
  try
    let m1' = incrementMove m1 in
    (m1', m2)
  with Overflow ->
    try (baseMove, incrementMove m2) with Overflow -> raise Overflow
;;

let rec incrementTurn (moves, order) =
  try
    let moves' = incrementMoves moves in
    (moves', order)
  with Overflow ->
    if order = maxOrder then raise Overflow
    else incrementTurn (baseMoves, maxOrder)
;;

let rec incrementConfig = function
    [] -> raise Overflow
  | head :: tail ->
      try (incrementTurn head) :: tail
      with Overflow -> baseTurn :: (incrementConfig tail)
;;

let rec stringOfConfig = function
    [] -> "[]"
  | ((d1, d2), order) :: tail ->
      let stringOfMove = function
	  N -> "N" | S -> "S" | E -> "E" | W -> "W"
      in
      (Printf.sprintf "((%s, %s), %d) :: " (stringOfMove d1) (stringOfMove d2) order)
	^ (stringOfConfig tail)
;;

type simResult =
    SimNormal
  | SimDeath of int
  | SimPushes of int
;;

let simulateMoves pos1 pos2 gameState ((m1, m2), order) =
  let simulateMove (x, y) = function
      N -> (x+1, y)
    | S -> (x-1, y)
    | E -> (x, y+1)
    | W -> (x, y-1)
  in
  let pos1' = ref (simulateMove pos1 m1) in
  let pos2' = ref (simulateMove pos2 m2) in
  if getType !pos1' gameState = Wall then pos1' := pos1;
  if getType !pos2' gameState = Wall then pos2' := pos2;
  let simulatePush pushes =
    let pos = !pos1' in
    let newpos = simulateMove pos (if pushes = 1 then m1 else m2) in
    if getType newpos gameState = Wall then (SimPushes(pushes), pos1, pos2)
    else if getType newpos gameState = Water then (SimDeath(3-pushes), pos, pos)
    else if pushes = 1 then (SimPushes(pushes), pos, newpos)
    else (SimPushes(pushes), newpos, pos)
  in
  if getType !pos1' gameState = Water then (SimDeath(1), !pos1', !pos2')
  else if getType !pos2' gameState = Water then (SimDeath(2), !pos1', !pos2')
  else if !pos1' = !pos2' then simulatePush (3-order)
  else if !pos1' = pos2 && order = 1 then
    begin
      pos2' := pos2;
      simulatePush 1
    end
  else if !pos2' = pos1 && order = 2 then
    begin
      pos1' := pos1;
      simulatePush 2
    end
  else
    (SimNormal, !pos1', !pos2')
;;

type simValues =
    {
     pushedValue : int;
     pushesValue : int;
     pushesIntoWaterValue : int;
     pushedIntoWaterValue : int;
     goalValue : int;
   };;

let offensiveSimValues =
  {
   pushedValue = -3;
   pushesValue = 3;
   pushesIntoWaterValue = 5;
   pushedIntoWaterValue = -10;
   goalValue = 5
 };;

let defensiveSimValues =
  {
   pushedValue = -5;
   pushesValue = -2;
   pushesIntoWaterValue = 1;
   pushedIntoWaterValue = -10;
   goalValue = 7
 };;

let rec assignConfigValue pos1 pos2 gameState cur values = function
    [] -> cur
  | moves :: tail ->
      let (result, pos1', pos2') = simulateMoves pos1 pos2 gameState moves in
      let newValue = match result with
	SimNormal -> cur
      |	SimDeath(1) -> cur + values.pushedIntoWaterValue
      |	SimDeath(_) -> cur + values.pushesIntoWaterValue
      |	SimPushes(1) -> cur + values.pushesValue
      |	SimPushes(_) -> cur + values.pushedValue
      in
      assignConfigValue pos1' pos2' gameState newValue values tail
;;

let findBestConfig pos1 pos2 gameState mode =
  let infinity = 100000 in
  let values = match mode with Offensive -> offensiveSimValues | Defensive -> defensiveSimValues in

  let config = ref baseConfig in
  let best = ref (-infinity) in
  let bestConfig = ref baseConfig in
  
  try
    while true do
      let value = assignConfigValue pos1 pos2 gameState 0 values !config in
      if value > !best then begin
	best := value;
	bestConfig := !config
      end;
      config := incrementConfig !config
    done;
    baseConfig
  with Overflow -> !bestConfig
;;

let runMoveStep (money: int) (pkgs: package list) (gameState: gameState):command =
  let position = robotLocation !myId gameState in
  let carried = robotPackages !myId gameState in
  let cweights = List.map (fun id -> packageWeight id position gameState) carried in
  let cweight = List.fold_left (fun cur w -> cur + w) 0 cweights in
  let mode = if cweight > !maxCap / 2 then Defensive else Offensive in
  let graph = computeGraph pkgs gameState mode in
  let info = findPath graph position in
  let dest = findGoal money pkgs gameState mode info in
  Printf.printf "GOAL = (%d, %d)\n" (fst dest) (snd dest);

  let nearby = findOthersAround gameState info 9 in
  let moveMode =
    if !turnsFighting < 0 then begin
      turnsFighting := 1 + !turnsFighting;
      Pathfinding
    end
    else if !turnsFighting > 5 then begin
      turnsFighting := -4;
      Pathfinding
    end
    else match nearby with
      None ->
	turnsFighting := 0;
	Pathfinding
    | Some(p) ->
	turnsFighting := 1 + !turnsFighting;
	Fighting(p)
  in
  match moveMode with
    Pathfinding ->
      currentBid := 1;
      print_endline "PATHFINDING";
      let path = find_path_path info dest in
      print_endline ("PATH " ^ (join (List.map (fun (a, b) -> Printf.sprintf "(%d, %d)" a b) path)));
      let (nx, ny) = findNextTile info dest in
      let (x, y) = position in
      if x = nx then
	if ny = y - 1 then (1, Move([W]))
	else if ny = y + 1 then (1, Move([E]))
	else (1, Drop([]))
      else
	if nx = x - 1 then (1, Move([S]))
	else (1, Move([N]))
  | Fighting(p) ->
      print_endline "FIGHTING";
      Printf.printf "BID = %d\n" !currentBid;
      let config = findBestConfig position p gameState mode in
      let ((dir, _), order) = List.hd config in
      let bid = if order = 1 then !currentBid else -(!currentBid) in
      currentBid := 2 * !currentBid;
      if !currentBid > money / 10 then currentBid := money / 10;
      (bid, Move([dir]))
;;

let runPickupStep (money: int) (pkgs: package list) (gameState: gameState):command =
  let position = robotLocation !myId gameState in
  let othersAround = checkOthersAround position gameState in
  let carried = robotPackages !myId gameState in
  let positionList = List.map (fun _ -> position) carried in
  let carriedPkgs = idsToPackages carried positionList gameState in
  match pickupStrategy othersAround position carriedPkgs pkgs !maxCap with
    None ->
      let result = runMoveStep money pkgs gameState in
      result
  | Some(cmd) -> cmd
;;

let runDropStep (money: int) (pkgs: package list) (gameState: gameState):command =
  let carried = robotPackages !myId gameState in
  let position = robotLocation !myId gameState in
  let folder cur id =
    let dest = packageDest id position gameState in
    if dest = position then id :: cur
    else cur
  in
  let toDrop = List.fold_left folder [] carried in
  if toDrop = [] then runPickupStep money pkgs gameState
  else (1, Drop(toDrop))
;;

let runStep (money: int) (pkgs: package list) (gameState: gameState) id cap:command =
  myId := id;
  maxCap := cap;
  let result = runDropStep money pkgs gameState in
  result
;;
